require "track"
require "camera"
IngameState = lust.Gamestate:extend()
local BufferedInput = require "buffered_input"

require "countdown_state"
require "race_over_state"
require "particlepool"

function startRace(trackData)
  local ig = IngameState(trackData)
  G.core.states:push(ig)
  G.core.states:push(CountdownState(ig))
end


function IngameState:spawnParticle(x, y, s, c, t, i, tq, g)
  local p = self.particlePool:create(x, y, s, c, t, i, tq, g)
  self.group:add(p)
  return p
end


function IngameState:new(trackData, replay)

  self.trackData = trackData
	self.group = lust.Group()
	self.world = love.physics.newWorld(0, 9.81 * meter, false)
  if love.getVersion() >= 11 then
    getmetatable(self.world).getContactList = getmetatable(self.world).getContacts
  end
	self.world:setCallbacks(function(a, b) self:collision(a,b) end)
	self.player = Player(self)
	self.group:add(self.player)
	self.track = Track(self, trackData)

	self.input = lust.BoundInput()
	self.input:bind_key("throttle", function() self:throttle(true) end, function() self:throttle(false) end)
	self.input:bind_key("break", function() self:brake(true) end, function() self.player:brake(false) end)
	self.input:bind_key("shift", function() if not self.countdown then self:shift() end end)
	self.input:bind_key("spin", function() if not self.countdown then self:spin() end end)
	self.input:bind_key("down", function() self:setPitch(1) end, function() self:setPitch(-1) end)
	self.input:bind_key("up", function() self:setPitch(-1) end, function() self:setPitch(1) end)
  self.input:bind_key("reset", function() 
    G.core.states:pop()
    startRace(self.trackData)
  end)

	self.input.action_map_func = function(key) return G.keyconfig:mapKeyToAction(key) end

  self.buffered_input = BufferedInput(self.input)
  if not G.headless then
    self.camera = Camera(self)
    self.camera:track(self.player)
  end
  self.time = 0
  self.countdown = true

  self.timeBacklog = 0

  if not G.headless then
    self.particlePool = ParticlePool(self)
  end

  if replay then
    self.isReplay = true
    self.replay = replay
    self.step = 1
    self.dummyInput = lust.BoundInput()
    self.dummyInput.action_map_func = function(key) return key end
  else
    self.replayKeys = {}
  end

  self.preTime = 200

  self.flux = flux.group()
end


function IngameState:newReplayState()
  if self.isReplay then
    local o = 3*(self.step-1)+1
    while self.time == self.replay[o] do
      --self.buffered_input[self.replay[o+1]](self.buffered_input, self.replay[o+2])
      table.insert(self.buffered_input.buffer,{self.replay[o+1], self.replay[o+2]})
      self.step = self.step + 1
      o = 3*(self.step-1)+1
    end
  else
    if not (self.win or self.player.dead) then
      local b = self.buffered_input.buffer
      if #b > 0 then 
        local t = self.time
        for i = 1, #b do
          table.insert(self.replayKeys, t)
          table.insert(self.replayKeys, b[i][1])
          table.insert(self.replayKeys, b[i][2])
        end
      end
    end
  end
end


function IngameState:throttle(onoff)
  self.player:throttle(onoff)
end


function IngameState:brake(onoff)
  self.player:brake(onoff)
end

function IngameState:shift()
  self.player:shift()
end

function IngameState:spin()
  self.player:spin()
end

function IngameState:setPitch(pitch)
  self.player:setPitch(pitch)
end


function IngameState:death()
  self.player:kill()
  if self.isReplay then
    if self.onQuitReplay then
      self.onQuitReplay()
    else
      G.core.states:pop()
    end
  else
    G.core.states:replace(RaceOverState(self))
  end

  if not G.headless then
    self.camera:track(nil)
  end
end


function IngameState:playerCollide(p, other)
  if self.win then
    return
  end

  if other:getCategory() == 1 then
    if not self.win then
      self:death()
      if G.headless then
        print(json.encode({status="dead"}))
        love.event.quit()
      end
    end
  elseif other:getCategory() == 16 then
    if G.headless then
      print(json.encode({status="win", time = self.time}))
      love.event.quit()
    end

    if not G.headless then
      G.sounds.finish:play()
    end
    self.win = true
    if self.player:shape() == 'plane' then
      self.player:shift()
    end

    if self.isReplay then
      if self.onQuitReplay then
        self.onQuitReplay()
      else
        G.core.states:pop()
      end
    else
      G.core.states:replace(RaceOverState(self))
    end
  end
end


function IngameState:collision(a, b)
	local cha = self.player.fixtures.airframe2
	local af = self.player.fixtures.airframe


  if self.player:shape() == 'car' then
    if a == cha then
      self:playerCollide(a, b)
    elseif b == cha then
      self:playerCollide(b, a)
    end
	end

  if self.player:shape() == 'plane' then
    if a == af then
      self:playerCollide(a, b)
    elseif b == af then
      self:playerCollide(b, a)
		end
	end

  if not G.headless then
    if a == self.player.fixtures.chassis or b == self.player.fixtures.chassis
      or a == self.player.fixtures.front_wheel or b == self.player.fixtures.front_wheel
      or a == self.player.fixtures.back_wheel or b == self.player.fixtures.back_wheel then
      G.sounds.thud:play()
    end
  end
end


function IngameState:enter()
  if self.isReplay then
    G.core.input:push(self.dummyInput)
  else
    G.core.input:push(self.buffered_input)
  end
  if not G.headless then
    music(true)
  end
end

function IngameState:leave()
	G.core.input:pop(self.buffered_input)
  if not G.headless then
    music(false)
  end
end


function IngameState:draw(dt)
  if G.headless then
    return
  end

  love.graphics.clear(127, 165, 228)
	self.camera:setup()
	self.track:drawPre(self.camera)
	self.group:draw(1-self.timeBacklog)
  self.particlePool:draw()
  self.track:drawPost()
  love.graphics.reset()

  local w, h = love.graphics.getDimensions()
  if not self.countdown and not self.win and not self.player.dead then
    self.track:drawMinimap(self.player)
    love.graphics.setColor(0,0,0,200)
    love.graphics.rectangle('fill', 0, 0, w, G.smallFont:getHeight() + 20)
    love.graphics.setColor(255,255,255,255)
    love.graphics.setFont(G.smallFont)
    love.graphics.print("Time: " .. self:format_time(), 10, 10)

    local v = lust.tools.length(self.player.bodies.chassis:getLinearVelocity())
    if self.player:shape() == 'car' then
      love.graphics.printf("Speed: ", 0, 10, w - 200, 'right')
      love.graphics.printf(string.format("%3d kph", v  / 3), 0, 10, w-10, 'right')
    else
      love.graphics.printf("Airspeed: ", 0, 10, w - 200, 'right')
      love.graphics.printf(string.format("%3d kts", v / (3*1.853)), 0, 10, w-10, 'right')

      if v < 30 * (3*1.853) then
        love.graphics.setColor(0,0,0,200)
        local l = G.smallFont:getWidth("Stall Warning")
        love.graphics.rectangle('fill', (w-(l+10))/2, 290, l+20, G.smallFont:getHeight()+20)

        love.graphics.setColor(255,255,255,255)
        love.graphics.printf("Stall Warning", 0, 300, w, 'center')
      end
    end
  end
end

function IngameState:format_time()
  local m = math.floor(self.time / 6000)
  local s = math.fmod(self.time / 100, 60)
  return string.format("%02d:%05.2f", m , s)

end


function IngameState:physicsUpdate(dt)
  if not self.countdown and not self.win and not self.player.dead then
    self.time = self.time + 1
  end
	self.group:update(dt)
	self.world:update(dt)
  if not G.headless then
    self.camera:update(dt)
  else
  end
end

local stepSize = 0.01 
function IngameState:update(dt)
  if self.preFrame then
    self.preFrame()
  end

  self.timeBacklog = self.timeBacklog + dt
  while self.timeBacklog >= stepSize do
    self.preTime = self.preTime - 1
    if self.preTime <= 0 then
      self.countdown = false
    end
    self:newReplayState()
    self.buffered_input:dispatch()
    self:physicsUpdate(stepSize)
    self.timeBacklog = self.timeBacklog - stepSize
  end

  self.track:update(dt)
	self.group:updateGfx(dt)
end
