local floorWidth = 80
Track = Object:extend()


function Track.load()
  Track.cloudImage = love.graphics.newImage("data/cloud.png")
  local w, h = Track.cloudImage:getDimensions()
  Track.cloudQuads = {
    love.graphics.newQuad(0.0,      0.0,      0.5 * w , 0.5 * h, w, h),
    love.graphics.newQuad(0.5 * w,  0.0,      0.5 * w , 0.5 * h, w, h),
    love.graphics.newQuad(0.0,      0.5 * h,  0.5 * w , 0.5 * h, w, h),
    love.graphics.newQuad(0.5 * w,  0.5 * h,  0.5 * w , 0.5 * h, w, h)
  }

  Track.mmShader = love.graphics.newShader [[
    vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
      return vec4(1.0, 1.0, 1.0, 1.0);
    }
  ]]

end


function Track:addCollider(c)
  if c.shape == "rectangle" then
    return self:addPolygon({
      c.x,              c.y,
      c.x,              c.y + c.height,
      c.x + c.width, c.y + c.height,
      c.x + c.width, c.y
    })
  elseif c.shape == "polygon" then
    local ps = {}
    for i, p in ipairs(c.polygon) do
      table.insert(ps, c.x + p.x)
      table.insert(ps, c.y + p.y)
    end
    return self:addPolygon(ps)
  end
end

renderScale = math.cos(math.rad(45))
function Track:addPolygon(ps)
  local ps2 = {}
  for i, p in ipairs(ps) do
    if i % 2 == 1 then
      table.insert(ps2, p)
    else
      table.insert(ps2, p * renderScale)
    end
  end

  local newDrawable = {points = ps2}

  local scaled = {}
  for i, p in ipairs(ps) do
    table.insert(scaled, p / meter)
  end
  newDrawable.body = love.physics.newBody(self.state.world, 0, 0, 'static')
  newDrawable.fixture = love.physics.newFixture(newDrawable.body, love.physics.newPolygonShape(ps), 1)
  newDrawable.fixture:setUserData('ground')

  return newDrawable
end


function Track:new(state, data, spawn)
  self.finishMinimap = {}
  self.pingT = 0
  local trackFile = data.track
  if not G.headless then
    self.arrow = G.core.images:load("data/arrow.png")
    self.finish = G.core.images:load("data/finish.png")
    self.finish:setWrap('repeat', 'repeat')

    self.texts = {}

    self.clouds = {}
    local w = trackFile.width * trackFile.tilewidth
    local h = trackFile.height * trackFile.tileheight
    self.tw = w
    self.th = h

    local count = math.floor(w*h / 800000)
    for i = 1, 3 do
      local s = 2^(i-1)
      local lcount = count / s
      self.clouds[i] = love.graphics.newSpriteBatch(Track.cloudImage, lcount)
      for j = 1, lcount do
        local x = love.math.random(1, w / s)
        local y = love.math.random(1, h / s)
        local q = love.math.random(1, 4)
        local sc = love.math.random() * 0.3 + 0.7

        sc = sc / s

        self.clouds[i]:add(Track.cloudQuads[q], x, y * renderScale, 0, sc, sc)
      end
    end

  end

  self.state = state

  self.colliders = {}

  self.preImages = {}
  self.postImages = {}
  self.layers = {}
  local fixImageLayerPath = function(p)
    return string.gsub(p, '../', 'data/')
  end
  for i, l in ipairs(trackFile.layers) do
    if l.type ~= "imagelayer" then
      self.layers[l.name] = l
    else
      local fname = fixImageLayerPath(l.image)
      local list
      if l.properties.order and l.properties.order == "after" then
        list = self.postImages
      else
        list = self.preImages
      end
      if not list[fname] then
        local img = G.core.images:load(fname)
        list[fname] = love.graphics.newSpriteBatch(img)
      end
      list[fname]:add(l.x, l.y * renderScale)
    end
  end

  self.walls = {verts={}, idcs={}}
  local floorColor = {250, 150, 50}
  if love.getVersion() >= 11 then
    floorColor = {250 / 255, 150 /255, 50/255}
  end
  for i, c in ipairs(self.layers.Track.objects) do
    local n = self:addCollider(c)
    n.fixture:setCategory(1)
    n.fixture:setFriction(1)
    table.insert(self.colliders, n)

    if not G.headless then
      local triangles = love.math.triangulate(n.points)
      local factor = 1
      if love.getVersion() >= 11 then
        factor = 1 / 255
      end
      for _, t in ipairs(triangles) do
        table.insert(self.walls.verts, {t[1], t[2] + renderScale*floorWidth, 0, 0, 150*factor, 70*factor, 20*factor})
        table.insert(self.walls.verts, {t[3], t[4] + renderScale*floorWidth, 0, 0, 150*factor, 70*factor, 20*factor})
        table.insert(self.walls.verts, {t[5], t[6] + renderScale*floorWidth, 0, 0, 150*factor, 70*factor, 20*factor})
      end
    end
  end

  if not G.headless then
    self.wallMesh = love.graphics.newMesh(self.walls.verts, "triangles", "static")

    self.arrows = {}
    self.finishLine = {verts = {}, idcs = {}}
  end
  for i, o in ipairs(self.layers.Entities.objects) do
    if o.type == "Player" then
      self.state.player:warpTo(o.x + o.width / 2, o.y + o.height / 2)
      if o.properties.scale then
        self.state.player.scale = tonumber(o.properties.scale)
      end

    elseif o.type == "Arrow" then
      table.insert(self.arrows, {x = o.x, y = o.y, rot = math.rad(o.rotation)})


    elseif not G.headless and o.type == "Text" then
      table.insert(self.texts, {
        text = o.properties["text"],
        x = o.x,
        y = o.y * renderScale
      })
    elseif o.type == "Finish" then
      local n = self:addCollider(o)
      table.insert(self.colliders, n)
      table.insert(self.finishMinimap, n.points)
      n.fixture:setCategory(16)
      n.fixture:setSensor(true)
    elseif not G.headless and o.type == "FinishLine" then
      local u2 = o.width /32 
      local v2 = o.height / 32

      local rot = math.rad(o.rotation)

      local p1 = {o.x, (o.y + floorWidth)* renderScale}
      local p2 = {lust.tools.rotate(o.width, 0, rot)}
      local p3 = {lust.tools.rotate(o.width, o.height, rot)}
      local p4 = {lust.tools.rotate(0, o.height, rot)}

      p2[1] = p2[1] + o.x
      p3[1] = p3[1] + o.x
      p4[1] = p4[1] + o.x
      p2[2] = p2[2] + o.y + floorWidth
      p3[2] = p3[2] + o.y+ floorWidth
      p4[2] = p4[2] + o.y+ floorWidth

      p2[2] = p2[2] * renderScale 
      p3[2] = p3[2] * renderScale 
      p4[2] = p4[2] * renderScale 

      p1[3] = 0
      p1[4] = 0
      p2[3] = u2
      p2[4] = 0
      p3[3] = u2
      p3[4] = v2
      p4[3] = 0
      p4[4] = v2

      table.insert(self.finishLine.verts, p1)
      table.insert(self.finishLine.verts, p2)
      table.insert(self.finishLine.verts, p3)
      table.insert(self.finishLine.verts, p4)

      table.insert(self.finishLine.idcs, #self.finishLine.verts-3)
      table.insert(self.finishLine.idcs, #self.finishLine.verts-2)
      table.insert(self.finishLine.idcs, #self.finishLine.verts)

      table.insert(self.finishLine.idcs, #self.finishLine.verts-1)
      table.insert(self.finishLine.idcs, #self.finishLine.verts)
      table.insert(self.finishLine.idcs, #self.finishLine.verts-2)
    end
  end

  if not G.headless and #self.finishLine.verts > 0 then
    self.finishMesh = love.graphics.newMesh(self.finishLine.verts, 'triangles', 'static')
    self.finishMesh:setVertexMap(self.finishLine.idcs)
    self.finishMesh:setTexture(self.finish)
  end

  if not G.headless then
    self.floors = {verts = {}, idcs={}}
    local floorColor = {250, 150, 50}
    if love.getVersion() >= 11 then
      floorColor = {250 / 255, 150 /255, 50/255}
    end
    for i, o in pairs(self.layers.Floors.objects) do
      if o.shape == "polyline" then
        for i = 1, #o.polyline-1 do
          self:add('floors', {x = o.x + o.polyline[i].x, y = o.y + o.polyline[i].y},
                             {x = o.x + o.polyline[i+1].x, y = o.y + o.polyline[i+1].y},
                             floorColor)
        end
      end
    end

    self.floorMesh = love.graphics.newMesh(self.floors.verts, "triangles", "static")
    self.floorMesh:setVertexMap(self.floors.idcs)
  end

  self:prepareMinimap()
end

function Track:add(what, p1, p2, c)
  local out = self[what].verts
  local outi= self[what].idcs

  local lightX, lightY = lust.tools.normalize(0.2, 1)
  local lx, ly = lust.tools.normalize(p2.x - p1.x, p2.y - p1.y)
  lx,ly = -ly, lx
  local swap = false
  if ly > 0 then
    ly = ly * -1
    lx = lx * -1
    swap = true
  end
  local l = math.min(
    math.max(0.05, -lightX*ly+lightY*lx),
    1.0
  )

  local r,g,b = c[1]*l, c[2]*l, c[3]*l

  local r2,g2,b2 = 0.7 * r, 0.7*g, 0.8*b

  if swap then
    r,g,b,r2,g2,b2 = r2,g2,b2,r,g,b
  end

  table.insert(out, { p1.x, renderScale*(p1.y - floorWidth), 0, 0, r2, g2, b2 })
  table.insert(out, { p1.x, renderScale*(p1.y + floorWidth), 0, 0, r, g, b })
  table.insert(out, { p2.x, renderScale*(p2.y + floorWidth), 0, 0, r, g, b })
  table.insert(out, { p2.x, renderScale*(p2.y - floorWidth), 0, 0, r2, g2, b2 })

  table.insert(outi, #out - 3)
  table.insert(outi, #out - 2)
  table.insert(outi, #out - 1)

  table.insert(outi, #out - 1)
  table.insert(outi, #out )
  table.insert(outi, #out - 3)

  
end


function Track:drawPost()
  if G.headless then
    return
  end

  love.graphics.setColor(255,255,255,255)
  love.graphics.draw(self.wallMesh)

  for _, i in pairs(self.postImages) do
    love.graphics.draw(i)
  end

  love.graphics.setFont(G.smallFont)
  for _, t in ipairs(self.texts) do
    love.graphics.setColor(0,0,0,200)
    local w = G.smallFont:getWidth(t.text)
    love.graphics.rectangle('fill', t.x - w/2 - 10, t.y -10, w + 20, G.smallFont:getHeight() + 20)
    love.graphics.setColor(255,255,255,255)
    love.graphics.print(t.text, t.x - w / 2, t.y)
  end

end


function Track:drawPre(cam)

  if G.headless then
    return
  end

  local cox = cam.x - self.tw / 2
  local coy = cam.y - self.th / 2
  local w, h = love.graphics.getDimensions()

  for i = 3, 1, -1 do
    love.graphics.push()
    love.graphics.origin()
    local s = 1/2^(i-0.75)
    love.graphics.translate(math.floor(-cam.x * s+ w / 2), math.floor(-cam.y * renderScale *s + h / 2))
    local l = 0.5 + 0.5 / i
    love.graphics.setColor(
      lust.tools.lerp(127, 255, l),
      lust.tools.lerp(165, 255, l),
      lust.tools.lerp(228, 255, l),
      255)
    love.graphics.draw(self.clouds[i])
    love.graphics.pop()
  end

  love.graphics.setColor(255,255,255,255)

  love.graphics.draw(self.floorMesh)

  for _, i in pairs(self.preImages) do
    love.graphics.draw(i)
  end

  if self.finishMesh then
    love.graphics.draw(self.finishMesh)
  end

  for _, a in ipairs(self.arrows) do
    love.graphics.draw(self.arrow, a.x, a.y * renderScale, a.rot, 1, 1, 32, 32)
  end
end


function Track:prepareMinimap()
  local aspect = self.tw / (self.th)
  local scale
  if aspect > 4/3 then
    scale = 200 / self.tw
  else
    scale = 150 / self.th
  end
  local w, h = love.graphics.getDimensions()
  self.minimapImg = love.graphics.newCanvas(self.tw * scale + 20, self.th * scale * renderScale + 20)
  love.graphics.setCanvas(self.minimapImg)
  love.graphics.clear(0,0,0,200)
  love.graphics.push()
  love.graphics.translate(10, 10)
  love.graphics.scale(scale, scale)
  love.graphics.setShader(Track.mmShader)
  love.graphics.draw(self.wallMesh)
  love.graphics.setShader()
  love.graphics.setCanvas()
  love.graphics.pop()
  self.minimapScale = scale
end


function Track:drawMinimap(player)
  local ww, wh = love.graphics.getDimensions()
  local w, h = self.minimapImg:getDimensions()
  local mms = self.minimapScale
  love.graphics.push()
  love.graphics.translate(ww - w, wh - h)
  love.graphics.draw(self.minimapImg)
  love.graphics.translate(10, 10)
  love.graphics.scale(mms, mms)
  local x, y = player:getPos()
  local r = ((6 * math.fmod(self.pingT*2, 1)) / mms) * Particle.scale 
  love.graphics.setColor(0,255,0,255 - math.fmod(self.pingT*2, 1) * 255)
  love.graphics.draw(Particle.img, x, y * renderScale, 0, r, r, Particle.offset, Particle.offset)
  love.graphics.setColor(255, 0, 0, 128 - math.sin(self.pingT*5)*32)
  for i, f in ipairs(self.finishMinimap) do
    love.graphics.polygon('fill', f)
  end

  love.graphics.pop()
end


function Track:update(dt)
  self.pingT = self.pingT + dt
end
